package it.unimi.dsi.webgraph;

/*		 
 * Copyright (C) 2003, 2004, 2005 Paolo Boldi and Sebastiano Vigna 
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.ints.AbstractIntIterator;
import it.unimi.dsi.fastutil.ints.IntArrays;
import it.unimi.dsi.fastutil.ints.IntIterator;
import it.unimi.dsi.fastutil.ints.IntIterators;
import it.unimi.dsi.fastutil.ints.IntSet;
import it.unimi.dsi.fastutil.io.BinIO;
import it.unimi.dsi.mg4j.util.MutableString;
import it.unimi.dsi.mg4j.util.ProgressMeter;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.NoSuchElementException;
import java.util.Properties;

/** An induced subgraph of a given immutable graph. 
 * 
 *  <P><strong>Warning</strong>: this class is experimental, and might be subject to changes.
 * 
 *  <P>The nodes of the subgraph are
 *  specified via an {@link it.unimi.dsi.fastutil.ints.IntSet} or an array of integers. Of course, each node in the subgraph will have
 *  a different index than the corresponding node in the supergraph. The two methods {@link #toSupergraphNode(int)} and {@link #fromSupergraphNode(int)}
 *  are used to translate indices back and forth.
 * 
 *  <P>An immutable subgraph is stored as a property file (which follows the convention established
 *  in {@link it.unimi.dsi.webgraph.ImmutableGraph}), and as a node subset file. The latter must contain an 
 *  (increasing) list of integers in {@link java.io.DataOutput} format representing
 *  the set of nodes of the subgraph.
 *
 *  <P>The property file, with named <samp><var>basename</var>.properties</samp>, contains the following entries:
 * 	<UL>
 *     <LI><samp>supergraphbasename</samp>: the basename of the supergraph. Note that this name is system-dependent:
 *		it is suggested that you use a path-free filename.
 *     <LI><samp>subgraphnodes</samp>: the filename of the node subset file. 
 * 	 If this property is not present, it is assumed to be <samp><var>basename</var>.subgraphnodes</samp>.
 * </UL>
 * 
 *   <P>You can create an immutable subgraph using the public constructor, or you can load one using one of the provided
 *   load methods. Note that there is no <code>store</code> method, because it makes no sense to store a generic {@link it.unimi.dsi.webgraph.ImmutableGraph}
 *   as a subgraph. There is, however, a {@linkplain #save(CharSequence, ProgressMeter) save method} that allows one to save
 *   the files related to a subgraph (the property file and the subgraph node file).
 * 
 * <H2>Root graphs</H2>
 * 
 * <P>When creating tree-shaped hierarchies of subgraphs, the methods {@link #rootBasename()}, {@link #fromRootNode(int)}
 * and {@link #toRootNode(int)} may be used to access information about the root (i.e., the unique highest graph
 * in the hierarchy: note that it cannot be an <code>ImmutableSubgraph</code>).
 * 
 * <P>Should you need to treat uniformly a generic immutable graph as an immutable subgraph, the method
 * {@link #asImmutableSubgraph(ImmutableGraph)} will return a subgraph view of the given immutable graph in which
 * all to/from functions are identities.
 */
public class ImmutableSubgraph extends ImmutableGraph {

	private final static boolean ASSERTS = false;
	private final static boolean DEBUG = false;

	private final static String THIS_CLASS_NAME = new ImmutableSubgraph( null ).getClass().getName();

	/** Builds an immutable subgraph from a given supergraph.
	 * 
	 * @param supergraph the supergraph. 
	 */
	private ImmutableSubgraph( final ImmutableGraph supergraph ) {
		this.supergraph = supergraph;
		this.supergraphAsSubgraph = supergraph instanceof ImmutableSubgraph ? (ImmutableSubgraph)supergraph : null;
	}
	
	/** The supergraph. */
	final protected ImmutableGraph supergraph;
	
	/** If {@link #supergraph} is an instance of {@link ImmutableSubgraph}, it is cached here. */
	final protected ImmutableSubgraph supergraphAsSubgraph;
	
	/** The nodes of the subgraph, in increasing order. */
	protected int subgraphNode[];

	/** The number of nodes in the subgraph. */
	protected int subgraphSize;

	/** The number of nodes in the supergraph. */
	protected int supergraphNumNodes;

	/** The basename of this immutable subgraph, if it was loaded from disk, or <code>null</code>. */
	protected CharSequence basename;

  	/** Creates a new immutable subgraph using a given subset of nodes.
  	 * 
  	 * @param supergraph the supegraph.
  	 * @param subgraphNodes the set of nodes defining the induced subgraph.
  	 */
	public ImmutableSubgraph( final ImmutableGraph supergraph, final IntSet subgraphNodes ) {
		this( supergraph );
		this.subgraphNode = subgraphNodes.toIntArray();
		this.subgraphSize = subgraphNode.length;
		this.supergraphNumNodes = supergraph.numNodes();
		java.util.Arrays.sort( subgraphNode );
		if ( subgraphNode[ subgraphSize - 1 ] >= supergraphNumNodes ) throw new IllegalArgumentException( "Subnode index out of bounds: " + subgraphNode[ subgraphSize - 1 ] );
	}

	/** Creates a new immutable subgraph using a given backing node array.
	 * 
	 * <P>Note that <code>subgraphNodes</code> is <em>not</em> copied: thus, it must not
	 * be modified until this subgraph is no longer in use.
	 * 
	 * @param supergraph the supegraph.
	 * @param subgraphNode an increasing array containing the nodes defining the induced subgraph. 
	 */
	public ImmutableSubgraph( final ImmutableGraph supergraph, final int subgraphNode[] ) {
		this( supergraph );
		this.subgraphNode = subgraphNode;
		this.subgraphSize = subgraphNode.length;
		this.supergraphNumNodes = supergraph.numNodes();
		for ( int i = 1; i < subgraphSize; i++ )
			if ( subgraphNode[ i - 1 ] >= subgraphNode[ i ] )
				throw new IllegalArgumentException( "The provided integer array is not strictly increasing: " + (i-1) + "-th element is " + subgraphNode[ i - 1 ] + ", " + i + "-th element is " + subgraphNode[ i ] );
		if ( subgraphNode[ subgraphSize - 1 ] >= supergraphNumNodes ) throw new IllegalArgumentException( "Subnode index out of bounds: " + subgraphNode[ subgraphSize - 1 ] );
	}

	public int numNodes() {
		return subgraphSize;
	}
	
	public long numArcs() {
		throw new UnsupportedOperationException( "Cannot determine the number of arcs in a subgraph" );
	}

	public CharSequence basename() {
		if ( basename == null ) throw new IllegalStateException( "This immutable subgraph has no basename" );
		return basename;
	}

	/** Returns the basename of the root graph.
	 * 
	 * @return the {@linkplain ImmutableGraph#basename() basename} of the root graph.
	 */
	public CharSequence rootBasename() {
		return supergraphAsSubgraph != null ? supergraphAsSubgraph.rootBasename() : supergraph.basename();
	}
	
	/** Returns the index of a node of this graph in its supergraph. 
	 * 
	 * @param x an index of a node in this graph.
	 * @return the index of node <code>x</code> in the supergraph.
	 */
	public int toSupergraphNode( final int x ) {
		if ( x < 0 || x >= subgraphSize ) throw new IllegalArgumentException();
		return subgraphNode[ x ];
	}

	/** Returns the index of a node of the supergraph in this graph. 
	 * 
	 * @param x an index of a node in the supergraph.
	 * @return the index of node <code>x</code> in this graph, or a negative value if <code>x</code> does not belong to the subgraph.
	 */
	public int fromSupergraphNode( final int x ) {
		if ( x < 0 || x >= supergraphNumNodes ) throw new IllegalArgumentException();
		return java.util.Arrays.binarySearch( subgraphNode, x );
	}
	
	/** Returns the index of a node of this graph in its root graph. 
	 * 
	 * @param x an index of a node in this graph.
	 * @return the index of node <code>x</code> in the root graph.
	 */
	public int toRootNode( final int x ) {
		return supergraphAsSubgraph != null ? supergraphAsSubgraph.toRootNode( toSupergraphNode( x ) ) : toSupergraphNode( x );
	}

	/** Returns the index of a node of the root graph in this graph. 
	 * 
	 * @param x an index of a node in the root graph.
	 * @return the index of node <code>x</code> in this graph, or a negative value if <code>x</code> does not belong to the root graph.
	 */
	public int fromRootNode( final int x ) {
		if ( supergraphAsSubgraph == null ) return fromSupergraphNode( x ); 
		final int y = supergraphAsSubgraph.fromRootNode( x ); 
		if ( y < 0 ) return -1;
		return fromSupergraphNode( y );
	}

	/** If this variable is non-negative, we are caching the successors' array of node <code>cacheNode</code> (in the subgraph). */
	private int cacheNode = -1;
	
	/** If <code>cacheNode</code>&gt; 0, this array contains the successors of node <code>cacheNode</code> (in the subgraph). */
	private int cacheSuccessors[];
	
	public NodeIterator nodeIterator( final int from ) {
		/** The invariant that we are assuming here is the following: at any time, <code>node</code> is the next (subgraph)
		 * node to be returned by {@link #nextInt()}. This variable contain sensible data
		 * only when <code>node</code> &lt; <code>subgraphSize</code>. Moreover, if outdegree >= 0 then it is 
		 * the outdegree of <code>node</code>-1, and <code>successorsCache</code> contains the successors. */

		// TODO: decide for a strategy
		return subgraphSize < supergraphNumNodes / 8 ? super.nodeIterator( from ) :

		new NodeIterator() {
				/** The current node (the next to be returned. */
				int node = from;
				/** This array caches the successors of the node that was returned last (<code>from</code>-1). */
				int[] successorsCache = IntArrays.EMPTY_ARRAY;
				/** The outdegree of the node that was returned last (<code>from</code>-1). */
				int outdegree = -1;
				
				final NodeIterator supergraphNodeIterator = supergraph.nodeIterator();

				public int nextInt() {
					if ( ! hasNext() ) throw new java.util.NoSuchElementException();
					if ( node == from ) {
						supergraphNodeIterator.nextInt();
						supergraphNodeIterator.skip( subgraphNode[ from ] );
					}
					else supergraphNodeIterator.skip( subgraphNode[ node ] - subgraphNode[ node - 1 ] );
					outdegree = -1;
					return node++;
				}

				public boolean hasNext() {
					return ( node < subgraphSize );
				}

				private void unwrapSuccessors() {
					int start = 0, done;
					final IntIterator i = ImmutableSubgraph.this.successors( node - 1, supergraphNodeIterator.successors() );
					while( ( done = IntIterators.unwrap( i, successorsCache, start, successorsCache.length - start ) ) == successorsCache.length - start && i.hasNext() ) { 
						start = successorsCache.length;
						successorsCache = IntArrays.grow( successorsCache, successorsCache.length + 1 ); 
					}
					outdegree = start + done; 
				}

				public int[] successorArray() {
					if ( node == from ) throw new IllegalStateException();
					if ( outdegree == -1 ) unwrapSuccessors();
					return successorsCache;
				}

				public IntIterator successors() {
					if ( node == from ) throw new IllegalStateException();
					if ( outdegree == -1 ) unwrapSuccessors();
					return IntIterators.wrap( successorsCache, 0, outdegree );
				}

				public int outdegree() {
					if ( node == from ) throw new IllegalStateException();
					if ( outdegree == -1 ) unwrapSuccessors();
					return outdegree;
				}
				
			};
	}
	
	public IntIterator successors( final int x ) {
		return successors( x, supergraph.successors( toSupergraphNode( x ) ) );
	}
	
	private IntIterator successors( final int x, final IntIterator supergraphSuccessors ) {
		if ( DEBUG ) System.err.println( this.getClass().getName() + ".successors(" + x + ", " + supergraphSuccessors + ")" );

		if ( x < 0 || x >= subgraphSize ) throw new IllegalArgumentException();
		if ( cacheNode == x ) return IntIterators.wrap( cacheSuccessors );
		
		if ( DEBUG ) System.err.println( this.getClass().getName() + ": returning new iterator" );

		return new AbstractIntIterator() {
			int cache = -1;
			
			public boolean hasNext() {
				if ( cache >= 0 ) return true;

				while ( supergraphSuccessors.hasNext() ) {
					cache = java.util.Arrays.binarySearch( subgraphNode, supergraphSuccessors.nextInt() );
					if ( cache >= 0 ) return true;
				}
				return false;
			}
			
			public int nextInt() {
				if ( !hasNext() ) throw new NoSuchElementException();
				int result = cache;
				cache = -1;
				return result;				
			}
		};
	}

	public int outdegree( final int x ) {
		return outdegree( x, supergraph.successors( toSupergraphNode( x ) ) );
	}

	public int outdegree( final int x, final IntIterator supergraphSuccessors ) {
		if ( x < 0 || x >= subgraphSize ) throw new IllegalArgumentException();
		if ( cacheNode == x ) return cacheSuccessors.length;
		cacheSuccessors = IntIterators.unwrap( successors( x, supergraphSuccessors ) );
		cacheNode = x;
		if ( ASSERTS ) assert cacheSuccessors != null;
		return cacheSuccessors.length;
	}

	/** A wrapper for immutable graphs, which exhibits them as immutable subgraphs.
	 * Essentially, all functions concerning supergraphs are defined as identities.
	 */
	
	private static class ImmutableGraphWrapper extends ImmutableSubgraph {
		
		public ImmutableGraphWrapper( final ImmutableGraph graph ) {
			super( graph );
			basename = graph.basename();
		}
		
		public NodeIterator nodeIterator() { return supergraph.nodeIterator(); }
		public NodeIterator nodeIterator( final int from ) { return supergraph.nodeIterator( from ); }
		public long numArcs() { return supergraph.numArcs(); }
		public int numNodes() { return supergraph.numNodes(); }
		public int outdegree( final int x ) { return supergraph.outdegree( x ); }
		public int[] successorArray( final int x ) { return supergraph.successorArray( x ); }
		public IntIterator successors( final int x ) { return supergraph.successors( x ); }
		public int toSupergraphNode( final int x ) { return x; }
		public int fromSupergraphNode( final int x ) { return x; }
		public int toRootNode( final int x ) { return x; }
		public int fromRootNode( final int x ) { return x; }
	}
	
	
	/** Returns a subgraph view of the given immutable graph.
	 * 
	 * <P>The wrapper returned by this method may be used whenever immutable
	 * graphs and subgraphs must be mixed.
	 * 
	 * @param graph an immutable graph.
	 * @return the given graph, viewed as a trivial subgraph of itself.
	 */
	public static ImmutableSubgraph asImmutableSubgraph( final ImmutableGraph graph ) {
		return new ImmutableGraphWrapper( graph );	
	}
	
	
	public static ImmutableGraph loadSequential( final CharSequence basename ) throws IOException {
		return load( "load", basename );
	}

	public static ImmutableGraph loadSequential( final CharSequence basename, final ProgressMeter pm ) throws IOException {
		return load( "load", basename, pm );
	}

	public static ImmutableGraph loadOffline( final CharSequence basename ) throws IOException {
		return load( "loadOffline", basename );
	}

	public static ImmutableGraph loadOffline( final CharSequence basename, final ProgressMeter pm ) throws IOException {
		return load( "loadOffline", basename, pm );
	}

	public static ImmutableGraph load( final CharSequence basename ) throws IOException {
		return load( "load", basename );
	}

	public static ImmutableGraph load( final CharSequence basename, final ProgressMeter pm) throws IOException {
		return load( "load", basename, pm );
	}

	private static ImmutableGraph load( final CharSequence method, final CharSequence basename ) throws IOException {
		return load( method, basename, null );
	}

	/** Creates a new immutable subgraph by loading the supergraph, delegating the 
	 *  actual loading to the class specified in the <samp>supergraphclass</samp> property within the property
	 *  file (named <samp><var>basename</var>.properties</samp>), and loading the subgraph array in memory.
	 *  The exact load method to be used depends on the <code>method</code> argument.
	 * 
	 * @param method the name of the method to be used to load the supergraph (<samp>load</samp>, <samp>loadSequential</samp>, <samp>loadOffline</samp>).
	 * @param basename the basename of the graph.
	 * @param pm the progress meter; it can be <code>null</code>.
	 * @return an immutable subgraph containing the specified graph.
	 */

	protected static ImmutableGraph load( final CharSequence method, final CharSequence basename, final ProgressMeter pm ) throws IOException {
		final FileInputStream propertyFile = new FileInputStream( basename + ".properties" );
		final Properties properties = new Properties();
		properties.load( propertyFile );
		propertyFile.close();

		final String graphClassName = properties.getProperty( "graphclass" );
		if ( ! graphClassName.equals( THIS_CLASS_NAME ) ) throw new IOException( "This class (" + THIS_CLASS_NAME + ") cannot load a graph stored using " + graphClassName );
		
		final String supergraphBasename = properties.getProperty( "supergraphbasename" );
		if ( supergraphBasename == null ) throw new IOException( "This property file does not specify the required property supergraphbasename" );
		
		final ImmutableGraph supergraph = ImmutableGraph.load( method, supergraphBasename, pm );
		
		if ( pm != null ) pm.out.print( "Reading nodes..." );
		final ImmutableSubgraph isg = new ImmutableSubgraph( supergraph, BinIO.loadInts( basename + ".nodes" ) );
		if ( pm != null ) pm.out.println( " done." );
		isg.basename = new MutableString( basename );
		return isg;
	}

	/** Throws an {@link UnsupportedOperationException}. */
	public static void store( final ImmutableGraph graph, final CharSequence basename, final ProgressMeter pm ) {
		if ( false && graph != null && basename != null && pm != null );
		throw new UnsupportedOperationException( "You cannot store a generic immutable graph as a subgraph" );
	}

	/** Throws an {@link UnsupportedOperationException}. */
	public static void store( final ImmutableGraph graph, final CharSequence basename ) {
		store( graph, basename, null );
	}
	
	/** Saves this immutable subgraph with a given basename. 
	 * 
	 * <P>Note that this method will <strong>not</strong> save the
	 * supergraph, but only the subgraph files, that is, the subgraph property file
	 * (with extension <samp>.properties</samp>) and the file containing
	 * the subgraph nodes (with extension <samp>.nodes</samp>). A reference
	 * to the supergraph basename will be stored in the property file.
	 * 
	 * @param basename the basename to be used to save the subgraph.
	 * @param pm a progress meter, or <code>null</code>.
	 */
	public void save( final CharSequence basename, final ProgressMeter pm ) throws IOException {

		final Properties properties = new Properties();
		properties.setProperty( "graphclass", THIS_CLASS_NAME );
		properties.setProperty( "supergraphbasename", supergraph.basename().toString() );

		final FileOutputStream propertyFile = new FileOutputStream( basename + ".properties" );
		properties.store( propertyFile, null );
		propertyFile.close();
		
		// Save the subgraph nodes
		if ( pm != null ) pm.start( "Saving nodes..." );
		BinIO.storeInts( subgraphNode, 0, subgraphNode.length, basename + ".nodes" );
		
		if ( pm != null ) {
			pm.count( subgraphNode.length );
			pm.done();
		}
	}

}
