package it.unimi.dsi.webgraph;


/*		 
 * Copyright (C) 2003, 2004, 2005 Paolo Boldi and Sebastiano Vigna 
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.ints.AbstractIntIterator;
import it.unimi.dsi.fastutil.ints.IntArrayList;
import it.unimi.dsi.fastutil.ints.IntIterator;
import it.unimi.dsi.fastutil.ints.IntLinkedOpenHashSet;

import java.util.NoSuchElementException;


/** A {@link MaskedIntIterator} is based on an array, the <em>mask</em>, and on an 
 *  <em>underlying iterator</em>. The sum of the values contained in the mask
 *  must not exceed the number of elements returned by the underlying iterator. Moreover, all integers in the mask
 *  must be positive, except possibly for the first one, which may be zero.
 *
 *  <P>Mask values are interpreted as specifying inclusion-exclusion blocks.
 *  Suppose that the underlying iterator returns <var>N</var> values, and that the mask is
 *  <var>n</var><sub>0</sub>, <var>n</var><sub>1</sub>, &hellip;, <var>n</var><sub>k</sub>.
 *  Then, the first <var>n</var><sub>0</sub> values returned by the underlying iterator must be kept,
 *  the next <var>n</var><sub>1</sub> values must be ignored, the next <var>n</var><sub>2</sub> must be
 *  kept and so on. The last <var>N</var>&minus;(<var>n</var><sub>0</sub>+&hellip;+<var>n</var><sub>k</sub>)
 *  must be kept if <var>k</var> is odd, and must be ignored otherwise.
 *  An instance of this class will returns the kept values only, in increasing order.
 */

public class MaskedIntIterator extends AbstractIntIterator {
	
	/** The underlying iterator. */
	private final IntIterator underlying;
	/** The mask. */
	private final int mask[];
	/** The mask. */
	private final int maskLen;
	/** This index in mask always represents an exclusion block. */
	private int currMask;
	/** How many integers are left in the current inclusion block. If <code>0</code> everything left must be discarded; if
	 *	<code>-1</code> all remaining values must be kept. */
	private int left;
	/** Whether {@link #underlying} is exhausted. This is accessed so frequently that it is better to cache it. */
	private boolean notOver;

	/** Creates a new masked iterator using a given mask and underlying iterator.
	 * 
	 * @param mask a mask, or <code>null</code>, meaning an empty mask (everything is copied).
	 * @param underlying an underlying iterator.
	 */


	public MaskedIntIterator( int mask[], IntIterator underlying ) {
		this( mask, mask == null ? 0 : mask.length, underlying );
	}

	/** Creates a new masked iterator using a given mask, mask length and underlying iterator.
	 * 
	 * @param mask a mask, or <code>null</code>, meaning an empty mask (everything is copied).
	 * @param maskLen an explicit mask length.
	 * @param underlying an underlying iterator.
	 */
	public MaskedIntIterator( final int mask[], final int maskLen, final IntIterator underlying ) {

		this.mask = mask;
		this.maskLen = maskLen;
		this.underlying = underlying;

		if ( currMask < maskLen ) {
			left = mask[ currMask++ ];
			advance();
		}
		else left = -1;
		notOver = underlying.hasNext();	
	}

	
	public int nextInt() throws NoSuchElementException {
		if ( ! hasNext() ) throw new NoSuchElementException( "Trying to get next from an empty MaskedIntIterator" );
		final int next = underlying.nextInt();
		if ( left > 0 ) {
			left--;
			advance();
		}
		notOver = underlying.hasNext();
		return next;
	}

	public boolean hasNext() {
		if ( left == 0 ) return false;
		if ( left < 0 ) return notOver;
		return true;
	}

	private void advance() {
		if ( left == 0 && currMask < maskLen ) {
			underlying.skip( mask[ currMask++ ] );
			if ( currMask < maskLen ) left = mask[ currMask++ ];
			else left = -1;	
		}
	}

	public int skip( int n ) {
		int skipped = 0;

		while( skipped < n && ( left > 0 || left == -1 && notOver ) ) {
			if ( left == -1 ) skipped += underlying.skip( n - skipped );
			else {
				if ( n - skipped < left ) {
					underlying.skip( n - skipped );
					notOver = underlying.hasNext();
					left -= ( n - skipped );
					return n;
				}
				else {
					underlying.skip( left );
					skipped += left;
					left = 0;
					advance();
					notOver = underlying.hasNext();
				}
			}
		}
			
		return skipped;
	}

	public static void main( String arg[] ) {
		// Reads the length and number of 0s
		int n = Integer.parseInt( arg[ 0 ] );
		int z = Integer.parseInt( arg[ 1 ] );
		final int x[] = new int[ n ];
		boolean keep[] = new boolean[ n ];
		IntArrayList res = new IntArrayList();
		IntArrayList blocks = new IntArrayList();
		int i, j, p = 0;
		boolean dep;

		// Generate
		for ( i = 0; i < n; i++ ) p = x[ i ] = p + (int)( Math.random() * 1000 );
		for ( i = 0; i < n-z; i++ ) keep[ i ] = true;
		for ( i = 0; i < n; i++ ) {
			j = i + (int)( Math.random() * ( n - i ) );
			dep = keep[ i ]; keep[ i ] = keep[ j ]; keep[ j ] = dep;
		}
		
		// Compute result
		for ( i = 0; i < n; i++ ) if ( keep[ i ] ) res.add( x[ i ] );
		res.trim();
		int results[] = res.elements();
		
		// Prepare blocks
		boolean lookAt = true;
		int curr = 0;
		for ( i = 0; i < n; i++ ) {
			if ( keep[ i ] == lookAt ) curr++;
			else {
				blocks.add( curr );
				lookAt = !lookAt;
				curr = 1;
			}
		}
		blocks.trim();
		final int bs[] = blocks.elements();

		// Output 
		System.out.println( "GENERATED:" );
		for ( i = 0; i < n; i++ ) {
			if ( keep[ i ] ) System.out.print( "*" );
			System.out.print( x[ i ] + "  " );
		}
		System.out.println( "\nBLOCKS:" );
		for ( i = 0; i < bs.length; i++ )
			System.out.print( bs[ i ] + "  " );
		System.out.println( "\nEXPECTED RESULT:" );
		for ( i = 0; i < results.length; i++ ) 
			System.out.print( results[ i ] + "  " );
		System.out.println();
		
		IntIterator ii = new MaskedIntIterator( bs, ( new IntLinkedOpenHashSet( x ) ).intIterator() );

		System.out.println( "OBTAINED RESULTS:" );
		int d, e;
		for ( i = 0; i < results.length; i++ ) {
			d = e = ii.nextInt(); // Tweak
			if ( e != d ) {
				System.out.println( "!!!!ERROR!!!! Current " + e + " does not match with next " + d );
			    return;
			}
			if ( results[ i ] != d ) {
				System.out.println( "!!!!ERROR!!!! I am going to return " + d + " instead of " + results[ i ] );
			    return;
			}
			System.out.print( d + "  " );
		}
		System.out.println();
		if ( ii.hasNext() ) System.out.println( "!!!!ERROR!!!! Iterator is not over yet; it would return " + ii.nextInt() );

		


	}


}

// Local Variables:
// mode: jde
// tab-width: 4
// End:
