package it.unimi.dsi.webgraph;

/*		 
 * Copyright (C) 2003, 2004, 2005 Paolo Boldi and Sebastiano Vigna 
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import it.unimi.dsi.fastutil.ints.AbstractIntIterator;
import it.unimi.dsi.fastutil.ints.IntAVLTreeSet;
import it.unimi.dsi.fastutil.ints.IntIterator;

import java.util.NoSuchElementException;

/** A merged integer iterator based on two {@link IntIterator}s that are supposed to return integers
 *  in an increasing fashion simply returns the (ordered) merge of the two iterators, without
 *  duplicates.
 */

public class MergedIntIterator extends AbstractIntIterator {
	
	private IntIterator it0, it1;
	private boolean valid0, valid1;
	private int curr0, curr1;
	private int n;

	public MergedIntIterator( IntIterator it0, IntIterator it1 ) {
		this ( it0, it1, Integer.MAX_VALUE );
	}

	/** Creates a new merged iterator by merging two given iterators; the resulting iterator will not emit more than <code>n</code> integers.
	 * 
	 * @param it0 the first increasing iterator.
	 * @param it1 the second increasing iterator.
	 * @param n the maximum number of integers this iterator will enumerate.
	 */
	public MergedIntIterator( IntIterator it0, IntIterator it1, int n ) {
		this.it0 = it0;
		this.it1 = it1;
		this.n = n;
		if ( valid0 = it0.hasNext() ) curr0 = it0.nextInt();
		if ( valid1 = it1.hasNext() ) curr1 = it1.nextInt();
	}

	public boolean hasNext() {
		return n != 0 && ( valid0 || valid1 );
	}

	public int nextInt() {
		if ( !hasNext() ) throw new NoSuchElementException( "Trying to get next from an empty MergedIntIterator" );
		n--;

		final int current;

		if ( !valid0 ) {
			current = curr1;
			if ( valid1 = it1.hasNext() ) curr1 = it1.nextInt();
			return current;
		} 
		if ( !valid1 ) {
			current = curr0;
			if ( valid0 = it0.hasNext() ) curr0 = it0.nextInt();
			return current;
		} 
		if ( curr0 < curr1 ) {
			current = curr0;
			if ( valid0 = it0.hasNext() ) curr0 = it0.nextInt();
		} 
		else if ( curr0 > curr1 ) {
			current = curr1;
			if ( valid1 = it1.hasNext() ) curr1 = it1.nextInt();
		} 
		else {
			current = curr0;
			if ( valid0 = it0.hasNext() ) curr0 = it0.nextInt();
			if ( valid1 = it1.hasNext() ) curr1 = it1.nextInt();
		}
		return current;
	}

	private static void error( String msg ) {
		System.err.println( "!!!!ERROR!!!!" + msg );
		System.exit( 1 );
	}

	public static void main( String arg[] ) {
		// Reads two lengths
		int n0 = Integer.parseInt( arg[ 0 ] );
		int n1 = Integer.parseInt( arg[ 1 ] );
		int x0[] = new int[ n0 ];
		int x1[] = new int[ n1 ];
		int i, p = 0;

		// Generate
		for ( i = 0; i < n0; i++ ) p = x0[ i ] = p + (int)( Math.random() * 10 );
		p = 0;
		for ( i = 0; i < n1; i++ ) p = x1[ i ] = p + (int)( Math.random() * 10 );
		
		IntAVLTreeSet s0 = new IntAVLTreeSet( x0 );
		IntAVLTreeSet s1 = new IntAVLTreeSet( x1 );
		IntAVLTreeSet res = new IntAVLTreeSet( s0 );
		res.addAll( s1 );

		System.out.println( "FIRST LIST:\n" + s0 );
		System.out.println( "SECOND LIST:\n" + s1 );
		System.out.println( "EXPECTED RESULT:\n" + res );
		
		IntIterator m = new MergedIntIterator( s0.intIterator(), s1.intIterator() );
		IntIterator it = res.intIterator();
		int x, y;

		while ( m.hasNext() ) {
			x = m.nextInt();
			//if ( m.currentInt() != x ) error( "Current is not sync'd" );
			y = it.nextInt();
			if ( x != y ) error( "Returning " + x + " instead of " + y );
		}
		if ( it.hasNext() ) error( "More elements should be returned (the first should be " + it.nextInt() + ")" );
				
	}
}


// Local Variables:
// mode: jde
// tab-width: 4
// End:
