package it.unimi.dsi.webgraph;

/*		 
 * Copyright (C) 2003, 2004, 2005 Paolo Boldi, Massimo Santini and Sebastiano Vigna 
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;
import it.unimi.dsi.fastutil.Iterators;
import it.unimi.dsi.fastutil.ints.AbstractIntIterator;
import it.unimi.dsi.fastutil.ints.IntArrays;
import it.unimi.dsi.fastutil.ints.IntIterator;
import it.unimi.dsi.fastutil.ints.IntIterators;
import it.unimi.dsi.fastutil.ints.IntOpenHashSet;
import it.unimi.dsi.fastutil.ints.IntSet;
import it.unimi.dsi.fastutil.io.FastBufferedInputStream;
import it.unimi.dsi.mg4j.util.ProgressMeter;
import jal.ints.BinaryPredicate;
import jal.ints.Inspection;
import jal.ints.Sorting;

import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.NoSuchElementException;

import cern.colt.function.IntComparator;

/** Static methods that manipulate immutable graphs. 
 * 
 *  <P>Most methods take an {@link
 *  it.unimi.dsi.webgraph.ImmutableGraph} (along with some other data, that
 *  depend on the kind of transformation), and return another {@link
 *  it.unimi.dsi.webgraph.ImmutableGraph} that represents the transformed
 *  version.
 */

public class Transform {
	
	private static final boolean DEBUG = false;
	private static final boolean ASSERTS = true;
	static { if ( DEBUG && ASSERTS ); }

	private final static String THIS_CLASS_NAME = new Transform().getClass().getName();
	
	private Transform() {}

	/** Provides a method to accept or reject an arc.
	 *
	 * <P>Note that arc filters are usually stateless. Thus, their declaration
	 * should comprise a static singleton (e.g., {@link Transform#NO_LOOPS}).
	 */
	public interface ArcFilter {
	
		/**
		 * Tells if the arc <code>(i,j)</code> has to be accepted or not.
		 * 
		 * @param i the source of the arc 
		 * @param j the destination of the arc
		 * @return if the arc has to be accepted
		 */
		public boolean accept( int i, int j );
	}

	/** An arc filter that rejects loops. */
	final static private class NoLoops implements ArcFilter {
		private NoLoops() {}
		
		public boolean accept( final int i, final int j ) {
			return i != j;
		}
	}

	/** A singleton providing an arc filter that rejects loops. */
	final static public NoLoops NO_LOOPS = new NoLoops();

	/** Returns the graph with some arcs eventually stripped, according to the given filter.
	 * 
	 * @param graph the graph.
	 * @param filter the filter (telling whether each arc should be kept or not).
	 * @param ignored a progress meter.
	 * @return the filtered-out graph.
	 */
	public static ImmutableGraph filterArcs( final ImmutableGraph graph, final ArcFilter filter, final ProgressMeter ignored ) {
		if ( false && ignored != null );
			
		return new ImmutableGraph() {
			private int succ[] = null;
			private int currentNode = -1;
	
			public int numNodes() { 
				return graph.numNodes(); 
			}
	
			public IntIterator successors( final int x ) {
				return new AbstractIntIterator() {
	
					private final IntIterator succ = graph.successors( x );
					private int cache = -1;
			
					public boolean hasNext() {				
						if ( cache >= 0 ) return true;
						else {
							while ( succ.hasNext() ) {
								int temp = succ.nextInt();
								if ( filter.accept( x, temp ) ) {
									cache = temp;
									break;
								} 
							}
							return cache >= 0;
						}
					}
			
					public int nextInt() {
						if ( !hasNext() ) throw new NoSuchElementException();
						int result = cache;
						cache = -1;
						return result;				
					}
	
				};
			}
			
			public int[] successorArray( int x ) {
				if ( currentNode != x ) {
					succ = IntIterators.unwrap( successors( x ) );
					currentNode = x ;
				}
				return succ;
			}
			
			public int outdegree( int x ) {
				if ( currentNode != x ) {
					succ = successorArray( x );
					currentNode = x;
				} 
				return succ.length;
			}
			
			public NodeIterator nodeIterator() {
				return new NodeIterator() {
					final NodeIterator nodeIterator = graph.nodeIterator();
					int[] succ = IntArrays.EMPTY_ARRAY;
					int outdegree = -1;

					public int outdegree() {
						if ( outdegree == -1 ) throw new IllegalStateException();
						return outdegree;
					}

					public int nextInt() {
						final int currNode = nodeIterator.nextInt();
						final int oldOutdegree = nodeIterator.outdegree();
						int[] oldSucc = nodeIterator.successorArray();
						succ = IntArrays.ensureCapacity( succ, oldOutdegree, 0 );
						outdegree = 0;
						for( int i = 0; i < oldOutdegree; i++ ) if ( filter.accept( currNode, oldSucc[ i ] ) ) succ[ outdegree++ ] = oldSucc[ i ];
						return currNode;
					}
					
					public int[] successorArray() {
						if ( outdegree == -1 ) throw new IllegalStateException();
						return succ;
					}

					public boolean hasNext() {
						return nodeIterator.hasNext();
					}
				};
			}
			
		};
	}
	
	/** Remaps the the graph nodes through a function specified via
	 *  an array. More specifically, <code>f.length=g.numNodes()</code>,
	 *  and <code>f[i]</code> is the new name of node <code>i</code>. (It is
	 *  required that the maximum integer in the image of <code>f</code>
	 *  is not greater than <code>f.length</code>.)
	 *  
	 *  <P>Arcs are mapped in the obvious way; in other words, there is
	 *  an arc from <code>f[i]</code> to <code>f[j]</code> in the transformed 
	 *  graph iff there was an arc from <code>i</code> to <code>j</code>
	 *  in the original graph. 
	 * 
	 *  <P>Note that if <code>f</code> is injective, the returned graph
	 *  is simply a permutation of the original graph. Otherwise,
	 *  the returned graph will be a quotient of the original graph
	 *  w.r.t. the equivalence relation induced by the fibres of <code>f</code>.
	 * 
	 *  <P>This method <strong>cannot</strong> process {@linkplain ImmutableGraph#loadOffline(CharSequence) offline
	 *  graphs}.
	 * 
	 * @param g the graph to be transformed.
	 * @param f the transformation map.
	 * @param pm a progress meter to be used during the precomputation (it can be <code>null</code>).
	 * @return the transformed graph.
	 */
	public static ImmutableGraph map( final ImmutableGraph g, final int f[], final ProgressMeter pm ) {	
		int i;
		final int sourceNumNodes = g.numNodes();
		if ( f.length != sourceNumNodes ) throw new IllegalArgumentException( "The graph to be mapped has " + sourceNumNodes + " whereas the map contains " + f.length + " entries" );

		// The pseudoinverse array
		final int sf[] = new int[ sourceNumNodes ];

		// sf[] = identity; compute the max image in f...
		int max = -1;
		if ( pm != null ) pm.start( "Storing identity..." );
		for ( i = 0; i < sourceNumNodes; i++ ) {
			if ( max < f[ i ] )	max = f[ i ];
			if ( f[ i ] < 0 ) throw new IllegalArgumentException( "The supplied map contains a negative value (" + f[ i ] +") at index " + i );
			sf[ i ] = i;
		}
		final int destNumNodes = max + 1;
		if ( destNumNodes > sourceNumNodes ) throw new IllegalArgumentException( "The map has a rank containing a value (" + max + ") not less than the number of nodes (" + sourceNumNodes + ")" );
		if ( pm != null ) {
			pm.count( sourceNumNodes );
			pm.itemsName( "nodes" );
			pm.stop( " done." ); 
			pm.out.println( pm );
		}

		// sort sf[]			
		if ( pm != null ) pm.start( "Sorting to obtain pseudoinverse..." );
		Sorting.sort( sf, 0, sourceNumNodes, new BinaryPredicate() {
			public boolean apply( int x, int y ) {
				return f[ x ] < f[ y ];
			}
		} );
		if ( pm != null ) {
			pm.count( sourceNumNodes );
			pm.stop( " done." ); 
			pm.out.println( pm );
		}
		
		// check if f is injective (by looking if there are consecutive entries of sf whose f value is the same)
		if ( pm != null ) pm.start( "Checking whether it is a permutation..." );
		final boolean isInjective = Inspection.adjacent_find( sf, 0, sourceNumNodes, new BinaryPredicate() {
				public boolean apply( int x, int y ) {
					return f[ x ] == f[ y ];
				}
			}) == sourceNumNodes;
		if ( pm != null ) {
			pm.count( sourceNumNodes );
			pm.stop( " done. (It is" + ( isInjective? "" : " not") + " a permutation)" ); 
			pm.out.println( pm );
		}

		return new ImmutableGraph() {
				
			private int[] succ = null;
			private int currentNode = -1;
				
			public int numNodes() {
				return destNumNodes;
			}
				
			public int[] successorArray( int x ) {
				if ( currentNode != x ) {
					IntSet succSet = new IntOpenHashSet();
					succSet.clear();
						
					if ( isInjective ) {
						IntIterator i = g.successors( sf[ x ] );
						while ( i.hasNext() ) succSet.add( f[ i.nextInt() ] );
					} 
					else {
						int low = 0, high = sourceNumNodes - 1, mid = 0;
						while ( low <= high ) {
							mid = ( low + high ) >> 1;
							int midVal = f[ sf[ mid ] ];
							if ( midVal < x )low = mid + 1;
							else if ( midVal > x )high = mid - 1;
							else break;
						}
						while ( mid > 0 && f[ sf[ mid - 1 ] ] == x ) mid--;
						int y = mid;
						while ( y < sourceNumNodes && f[ sf[ y ] ] == x ) {
							IntIterator i = g.successors( sf[ y++ ] );
							while ( i.hasNext() ) succSet.add( f[ i.nextInt() ] );
						}
					}
					succ = succSet.toIntArray();
					if ( succ.length > 0 ) Arrays.sort( succ, 0, succ.length );
					currentNode = x;
				}
				return succ;
			}
				
			public int outdegree( int x ) {
				if ( currentNode == x ) return succ.length;
				return successorArray( x ).length;
			}
				
		};
	}

	/** Remaps the the graph nodes through a function specified via
	 *  an array.
	 * 
	 * @param g the graph to be transformed.
	 * @param f the transformation map.
	 * @return the transformed graph.
	 * @see #map(ImmutableGraph, int[], ProgressMeter)
	 */
	public static ImmutableGraph map( final ImmutableGraph g, final int f[] ) {
		return map( g, f, null );
	}

	/** Returns a symmetrized graph.
	 * 
	 * <P>The symmetrized graph is the union of a graph and of its transpose. This method will
	 * use the provided transposed graph, if any, instead of computing it on-the-fly.  
	 * 
	 * @param g the source graph.
	 * @param t the graph g transposed; if <code>null</code>, the transposed graph will be computed on the fly.
	 * @param pm a progress meter.
	 * @return the symmetryzed graph.
	 */
	public static ImmutableGraph symmetrize( final ImmutableGraph g, final ImmutableGraph t, final ProgressMeter pm ) {
		return t == null ?
			union( g, it.unimi.dsi.webgraph.Transform.transpose( g, pm ), pm ) :
			union( g, t, pm );
	}

	/** Returns a symmetrized graph.
	 * 
	 * @param g the source graph.
	 * @param pm a progress meter.
	 * @return the symmetryzed graph.
	 * @see #symmetrize(ImmutableGraph, ImmutableGraph, ProgressMeter)
	 */
	public static ImmutableGraph symmetrize( final ImmutableGraph g, final ProgressMeter pm ) {
		return symmetrize( g, null, pm );
	}

	/** Returns a symmetrized graph.
	 * 
	 * @param g the source graph.
	 * @return the symmetryzed graph.
	 * @see #symmetrize(ImmutableGraph, ImmutableGraph, ProgressMeter)
	 */
	public static ImmutableGraph symmetrize( final ImmutableGraph g ) {
		return symmetrize( g, null );
	}

	/** Returns an immutable graph obtained by reversing all arcs in <code>g</code>.
	 * 
	 * <P>This method can process {@linkplain ImmutableGraph#loadOffline(CharSequence) offline graphs}).
	 * 
	 * @param g an immutable graph.
	 * @param pm a progress meter.
	 * @return an immutable graph obtained by transposing <code>g</code>.
	 */	
	
	public static ImmutableGraph transpose( final ImmutableGraph g, final ProgressMeter pm ) {
	
		int i, j, d, a[];
	
		final int n = g.numNodes();
		final int numPred[] = new int[ n ];
	
		if ( pm != null ) {
			pm.out.print( "Counting predecessors..." ); 
			pm.itemsName( "nodes" );
			pm.start();
		}
		
		NodeIterator nodeIterator = g.nodeIterator();
		i = n;
		while( i-- != 0 ) {
			nodeIterator.nextInt();
			d = nodeIterator.outdegree();
			a = nodeIterator.successorArray();
			while( d-- != 0 ) numPred[ a[ d ] ]++;
			if ( pm != null )pm.update();
		}
	
		if ( pm != null ) {
			pm.stop();
			pm.out.println( " done." );
			pm.out.println( pm );
		}
	
		final int pred[][] = new int[ n ][];
	
		if ( pm != null ) pm.out.print( "Allocating memory for predecessors..." );
		i = n;
		while( i-- != 0 ) if ( numPred[ i ] != 0 ) pred[ i ] = new int[ numPred[ i ] ];
		if ( pm != null ) pm.out.println( " done." );
	
		i = n;
		while( i-- != 0 ) numPred[ i ] = 0;
	
		if ( pm != null ) {
			pm.out.print( "Computing predecessors..." );
			pm.start();
		}
	
		nodeIterator = g.nodeIterator();
		i = n;
		
		while( i-- != 0 ) {
			j = nodeIterator.nextInt();
			d = nodeIterator.outdegree();
			a = nodeIterator.successorArray();
			while( d-- != 0 ) pred[ a[ d ] ][ numPred[ a[ d ] ]++ ] = j;
			if ( pm != null ) pm.update();
		}
		if ( pm != null ) {
			pm.stop();
			pm.out.println( " done." ); 
			pm.out.println( pm );
		}
	
		i = n;
		if ( pm != null ) { 
			pm.out.print( "Sorting predecessors..." );
			pm.start();
		}
		while( i-- != 0 ) {
			if ( pred[ i ] != null ) Arrays.sort( pred[ i ] );
			if ( pm != null ) pm.update();
		} 
		
		if ( pm != null ) {
			pm.stop();
			pm.out.println(" done.");
			pm.out.println(pm);
		}
	
		return new ImmutableGraph() {
			public int numNodes() { return n; }
			public int[] successorArray( int x ) { return pred[ x ] != null ? pred[ x ] : IntArrays.EMPTY_ARRAY; }
			public int outdegree( int x ) { return numPred[ x ]; }
		};
	}	
	
	
	/** Returns an immutable graph obtained by reversing all arcs in <code>g</code>.
	 * 
	 * <P>This method can process {@linkplain ImmutableGraph#loadOffline(CharSequence) offline graphs}).
	 * 
	 * @param g an immutable graph.
	 * @return an immutable graph obtained by transposing <code>g</code>.
	 * @see #transpose(ImmutableGraph, ProgressMeter)
	 */	
	
	public static ImmutableGraph transpose( ImmutableGraph g ) {
		return transpose( g, null );
	}

	/** Returns the union of two immutable graphs.
	 * 
	 * <P>The two arguments may differ in the number of nodes, in which case the
	 * resulting graph will be large as the larger graph.
	 * 
	 * @param g0 the first graph.
	 * @param g1 the second graph.
	 * @param ignored an unused progress meter.
	 * @return the union of the two graphs.
	 */
	public static ImmutableGraph union( final ImmutableGraph g0, final ImmutableGraph g1, final ProgressMeter ignored ) {
		if ( false && ignored != null );
		
		return new ImmutableGraph() {
	
			private int n0 = g0.numNodes();
			private int n1 = g1.numNodes();
			final int numNodes = Math.max( n0, n1 );
			/** The node whose successors are cached, or -1 if no successors are currently cached. */
			private int cachedNode = -1;
			/** The outdegree of the cached node, if any. */
			private int outdegree ;
			/** The successors of the cached node, if any; note that the array might be larger. */
			private int cache[] = IntArrays.EMPTY_ARRAY;
	
			public NodeIterator nodeIterator( final int from ) {
	
				return new NodeIterator() {
					/** If outdegree is nonnegative, the successors of the current node (this array may be, however, larger). */
					private int cache[] = IntArrays.EMPTY_ARRAY;
					/** The outdegree of the current node, or -1 if the successor array for the current node has not been computed yet. */
					private int outdegree = -1;
					private NodeIterator i0 = from < n0? g0.nodeIterator( from ) : null;
					private NodeIterator i1 = from < n1? g1.nodeIterator( from ) : null;

					public boolean hasNext() {
						return i0 != null && i0.hasNext() || i1 != null && i1.hasNext();
					}
					
					public int nextInt() {
						if ( ! hasNext() ) throw new java.util.NoSuchElementException();
						outdegree = -1;
						int result = -1;
						if ( i0 != null ) {
							if ( i0.hasNext() ) result = i0.nextInt();
							else i0 = null;
						}
						if ( i1 != null ) {
							if ( i1.hasNext() ) result = i1.nextInt();
							else i1 = null;
						}
						return result;
					}
	
					public int[] successorArray() {
						if ( outdegree != -1 ) return cache;
						if ( i0 == null ) {
							outdegree = i1.outdegree();
							return cache = i1.successorArray();
						} 
						if ( i1 == null ) {
							outdegree = i0.outdegree();
							return cache = i0.successorArray();
						}
											
						IntIterator merge = new MergedIntIterator( i0.successors(), i1.successors() );
						outdegree = 0;
						outdegree += IntIterators.unwrap( merge, cache );
						int upto;
						while ( merge.hasNext() ) {
							upto = cache.length;
							cache = IntArrays.grow( cache, upto + 1 );
							outdegree += IntIterators.unwrap( merge, cache, upto, cache.length - upto );
						}
						return cache;
					}
	
					public int outdegree() {
						successorArray(); // So that the cache is filled up
						return outdegree;
					}
	                            
				};
			
			}
	
			public int numNodes() {
				return numNodes;
			}
	
			public int[] successorArray( int x ) {
				if ( x == cachedNode ) return cache; 
				IntIterator merge = new MergedIntIterator( x < n0? g0.successors( x ) : Iterators.EMPTY_ITERATOR, x < n1? g1.successors( x ) : Iterators.EMPTY_ITERATOR );
				outdegree = 0;
				outdegree += IntIterators.unwrap( merge, cache );
				int upto;
				while ( merge.hasNext() ) {
					upto = cache.length;
					cache = IntArrays.grow( cache, upto + 1 );
					outdegree += IntIterators.unwrap( merge, cache, upto, cache.length - upto );
				}
				return cache;
			}
		
			public int outdegree( int x ) {
				successorArray( x ); // So the cache gets filled
				return outdegree;
			}
			
		};
	}

	/** Returns the union of two immutable graphs.
	 * 
	 * @param g0 the first graph.
	 * @param g1 the second graph.
	 * @return the union of the two graphs.
	 * @see #union(ImmutableGraph, ImmutableGraph, ProgressMeter)	 
	 */
	public static ImmutableGraph union( final ImmutableGraph g0, final ImmutableGraph g1 ) {
		return union( g0, g1, null );
	}


	/** Returns a permutation that would make the given graph adjacency lists in Gray-code order.
	 * 
	 * <P>Gray codes list all sequences of <var>n</var> zeros and ones in such a way that
	 * adjacent lists differ by exactly one bit. If we assign to each row of the adjacency matrix of 
	 * a graph its index as a Gray code, we obtain a permutation that will make similar lines
	 * nearer.
	 * 
	 * <P>Note that since a graph permutation permutes <em>both</em> rows and columns, this transformation is
	 * not idempotent: the Gray-code permutation produced from a matrix that has been Gray-code ordered will
	 * <em>not</em> be, in general, the identity.
	 * 
	 * <P>The important feature of Gray-code ordering is that it is completely endogenous (e.g., determined
	 * by the graph itself), contrarily to, say, lexicographic URL ordering (which relies on the knowledge
	 * of the URL associated to each node).
	 * 
	 * @param g an immutable graph.
	 * @return the permutation that would order the graph adjacency lists by Gray order
	 * (you can just pass it to {@link #map(ImmutableGraph, int[], ProgressMeter)}).
	 */ 
	public static int[] grayCodePermutation( final ImmutableGraph g ) {
		final int n = g.numNodes();
		final int[] perm = new int[ n ];
		int i = n;
		while( i-- != 0 ) perm[ i ] = i;
		
		final IntComparator grayComparator =  new IntComparator() {
			/* Remember that given a Gray code G (expressed as a 0-based sequence
			 * of n bits G[i]), the corresponding binary code B if defined as
			 * follows: B[n-1]=G[n-1], and B[i] = B[i+1] ^ G[i].
			 *
			 * Translating the formula above to our case (where we just have the increasing
			 * list of indices j such that G[i]=1), we see that the binary code
			 * corresponding to the Gray code of an adjacency list is 
			 * made of alternating blocks of zeroes and ones; the alternation
			 * happens at each successor. 
			 * 
			 * Said that, the code below requires some reckoning to be fully 
			 * understood (but it works!).
			 */

			public int compare( final int x, final int y ) {
				final IntIterator i = g.successors( x ), j = g.successors( y );
				int a, b;

				/* This code duplicates eagerly of the behaviour of the lazy comparator
				   below. It is here for documentation and debugging purposes.
				
				byte[] g1 = new byte[ g.numNodes() ], g2 = new byte[ g.numNodes() ];
				while( i.hasNext() ) g1[ g.numNodes() - 1 - i.nextInt() ] = 1;
				while( j.hasNext() ) g2[ g.numNodes() - 1 - j.nextInt() ] = 1;
				for( int k = g.numNodes() - 2; k >= 0; k-- ) {
					g1[ k ] ^= g1[ k + 1 ];
					g2[ k ] ^= g2[ k + 1 ];
				}
				for( int k = g.numNodes() - 1; k >= 0; k-- ) if ( g1[ k ] != g2[ k ] ) return g1[ k ] - g2[ k ];
				return 0;
				*/
				
				boolean lastBit = false;
				for( ;; ) {
					if ( ! i.hasNext() && ! j.hasNext() ) return 0;
					if ( ! i.hasNext() ) return lastBit ? 1 : -1;
					if ( ! j.hasNext() ) return lastBit ? -1 : 1;
					a = i.nextInt();
					b = j.nextInt();
					if ( a != b ) return lastBit ^ ( a < b ) ? 1 : -1;
					lastBit = ! lastBit;
				}				
			}
		};
		
		cern.colt.Sorting.quickSort( perm, 0, n, grayComparator );
		
		if ( ASSERTS ) for( int k = 0; k < n - 1; k++ ) assert grayComparator.compare( perm[ k ], perm[ k + 1 ] ) <= 0;  
		
		final int[] invPerm = new int[ n ];
		i = n;
		while( i-- != 0 ) invPerm[ perm[ i ] ] = i;
		
		return invPerm;
	}


	
	/** Ensures that the arguments are exactly <code>n</code>, if <code>n</code> is nonnegative, or
	 * at least -<code>n</code>, otherwise.
	 */

	private static boolean ensureNumArgs( Getopt g, String args[], int n ) {
		if ( n >= 0 && args.length - g.getOptind() != n || n < 0 && args.length - g.getOptind() < -n ) {
			System.err.println( "Wrong number (" + ( args.length - g.getOptind() ) + ") of arguments." );
			return false;
		}
		return true;
	}

	/** Loads a graph with given data and returns it.
	 *  
	 * @param graphClass the class of the graph to be loaded.
	 * @param baseName the graph basename. 
	 * @param sequential whether the graph is to be loaded in a sequential fashion.
	 * @param offline whether the graph is to be loaded in an offline fashion.
	 * @param pm a progress meter.
	 * @return the loaded graph.
	 */
	private static ImmutableGraph load( Class graphClass, String baseName, boolean sequential, boolean offline, ProgressMeter pm ) throws IllegalArgumentException, SecurityException, IllegalAccessException, InvocationTargetException, NoSuchMethodException, IOException {	
		ImmutableGraph graph = null;

		if ( graphClass != null ) {
			if ( offline ) graph = (ImmutableGraph)graphClass.getMethod( "loadOffline", new Class[] { CharSequence.class } ).invoke( graphClass, new Object[] { baseName } );
			else if ( sequential ) graph = (ImmutableGraph)graphClass.getMethod( "loadSequential", new Class[] { CharSequence.class, ProgressMeter.class } ).invoke( graphClass, new Object[] { baseName, pm } );
			else graph = (ImmutableGraph)graphClass.getMethod( "load", new Class[] { CharSequence.class, ProgressMeter.class } ).invoke( graphClass, new Object[] { baseName, pm } );
		}
		else graph = offline ? 
			ImmutableGraph.loadOffline( baseName ) : 
			sequential ? ImmutableGraph.loadSequential( baseName, pm ) :
			ImmutableGraph.load( baseName, pm );
		
		return graph;
	}


	public static void main( String args[] ) throws IOException, IllegalArgumentException, SecurityException, InstantiationException, IllegalAccessException, InvocationTargetException, NoSuchMethodException, ClassNotFoundException {
		Class sourceGraphClass = null, destGraphClass = BVGraph.class;
		boolean offline = false, sequential = false, ascii = false;
		int quantum = 100000;

		int ch;
		LongOpt longopts[] = new LongOpt[] {
			new LongOpt( "help", LongOpt.NO_ARGUMENT, null, 'h' ),
			new LongOpt( "source-graph-class", LongOpt.REQUIRED_ARGUMENT, null, 's' ),
			new LongOpt( "dest-graph-class", LongOpt.REQUIRED_ARGUMENT, null, 'd' ),
			new LongOpt( "quantum", LongOpt.REQUIRED_ARGUMENT, null, 'q' ),
			new LongOpt( "offline", LongOpt.NO_ARGUMENT, null, 'o' ),
			new LongOpt( "sequential", LongOpt.NO_ARGUMENT, null, 's' ),
			new LongOpt( "ascii", LongOpt.NO_ARGUMENT, null, 's' ),
		};

		Getopt g = new Getopt( "Transform", args, "q:hosc:S:D:a", longopts );
		g.setOpterr( true );

		while ( ( ch = g.getopt() ) != -1 ) switch ( ch ) {
		case 'h':
			System.err.println( "Usage: " + THIS_CLASS_NAME + "[OPTIONS] TRANSFORM SOURCE+ DEST [PARAMS]" );
			System.err.println( "Transforms one or more graphs. The SOURCE and DEST parameters are" );
			System.err.println( "basenames from which suitable file names will be stemmed." );
			System.err.println( "" );
			System.err.println( "Available transforms:" );
			System.err.println( "  identity, map, transpose   one source graph only" );
			System.err.println( "  symmetrize                 one or two source graphs" );
			System.err.println( "  union                      two source graphs" );
			System.err.println( "  gray                       one source graph" );
			System.err.println( "  arcfilter ARCFILTER        filters out arcs with the given arc filter" );
			final Field[] field = Transform.class.getDeclaredFields();
			final List filterList = new ArrayList();
			for( int i = 0; i < field.length; i++ ) if ( Arrays.asList( field[ i ].getType().getInterfaces() ).indexOf( ArcFilter.class ) != -1 ) filterList.add( field[ i ].getName() );
			System.err.println( "                             available filters: " + filterList );
			
			System.err.println( "" );
			System.err.println( "Optional arguments:" );
			System.err.println( "  -S, --source-graph-class   forces a Java class to load the source graph" );
			System.err.println( "  -D, --dest-graph-class     forces a Java class to load the source graph" );
			System.err.println( "  -o, --offline              use the offline load method to reduce memory consumption" );
			System.err.println( "  -s, --sequential           use the sequential load method to reduce memory consumption" );
			System.err.println( "  -q, --quantum              the progress meter quantum (default: " + quantum + ")" );
			System.err.println( "  -a, --ascii                the map is in ascii form (one int per line)" );			
			System.err.println( "" );
			System.err.println( "Help:" );
			System.err.println( "  -h, --help                 print this help screen" );
			System.err.println( "" );
			return;

		case 'S':
			sourceGraphClass = Class.forName( g.getOptarg(), true, ClassLoader.getSystemClassLoader() );
			break;

		case 'D':
			destGraphClass = Class.forName( g.getOptarg(), true, ClassLoader.getSystemClassLoader() );
			break;

		case 'o':
			offline = true;
			break;

		case 's':
			sequential = true;
			break;

		case 'a':
			ascii = true;
			break;

		case 'q':
			quantum = Integer.parseInt( g.getOptarg() );
			break;

		default:
			break;
		}

		if ( !ensureNumArgs( g, args, -3 ) ) return;
		
		String transform = args[ g.getOptind() ];
		
		String source[] = null, dest = null, map = null;
		ArcFilter arcFilter = null;
		
		final int optInd = g.getOptind();
		
		if ( transform.equals( "identity" ) || transform.equals( "transpose" ) || transform.equals( "map" ) || transform.equals( "gray" ) ) {
			source = new String[] { args[ optInd + 1 ] };
			dest = args[ optInd + 2 ];
			if ( transform.equals( "map" ) ) {
				if ( ! ensureNumArgs( g, args, 4 ) ) return;
				map = args[ g.getOptind() + 3 ];
			} 
			else if ( ! ensureNumArgs( g, args, 3 ) ) return;
		} 
		else if ( transform.equals( "symmetrize" ) ) {
			if ( ensureNumArgs( g, args, 3 ) ) {
				source = new String[] { args[ optInd + 1 ], null };
				dest = args[ optInd + 2 ];
			} 
			else if ( ensureNumArgs( g, args, 4 ) ) {
				source = new String[] { args[ optInd + 1 ], args[ optInd + 2 ] };
				dest = args[ optInd + 3 ];
			}
			else return;
		} 
		else if ( transform.equals( "arcfilter" ) ) {
			if ( ensureNumArgs( g, args, 4 ) ) {
				try {
					// First try: a public field
					arcFilter = (ArcFilter) Transform.class.getField( args[ optInd + 1 ] ).get( null );
				}
				catch( NoSuchFieldException e ) {
					// No chance: let's try with a class
					arcFilter = (ArcFilter) Class.forName( args[ optInd + 1 ] ).getConstructor( new Class[] {} ).newInstance( new Object[] {} );
				}
				source = new String[] { args[ optInd + 2 ], null };
				dest = args[ optInd + 3 ];
			} 
			else return;
		} 
		else if ( transform.equals( "union" ) ) {
			if ( ! ensureNumArgs( g, args, 4 ) ) return;
			source = new String[] { args[ optInd + 1 ], args[ optInd + 2 ] };
			dest = args[ optInd + 3 ];
		}
		else {
			System.err.println( "Unknown transform: " + transform );
			return;
		}

		final ProgressMeter pm = new ProgressMeter();
		pm.quantum( quantum );
		final ImmutableGraph graph[] = new ImmutableGraph[ source.length ];
		
		for ( int i = 0; i < source.length; i++ ) 
			if ( source[ i ] == null ) graph[ i ] = null;
			else graph[ i ] = load( sourceGraphClass, source[ i ], sequential, offline, pm );
			
		if ( transform.equals( "identity" ) ) ImmutableGraph.store( destGraphClass, graph[ 0 ], dest,  pm );
		else if ( transform.equals( "map" ) ) {
			System.err.print( "Reading function..." ); 
			final int n = graph[ 0 ].numNodes();
			final int[] f = new int[ n ];
			if ( ascii ) {
				final BufferedReader br = new BufferedReader( map.equals( "-" ) ? new InputStreamReader( System.in ) : new FileReader( map ), 1024 * 1024 );
				for( int i = 0 ; i < f.length; i++ ) f[ i ] = Integer.parseInt( br.readLine() );   
				br.close();
			}
			else {
				final DataInputStream dis = new DataInputStream( new FastBufferedInputStream( map.equals( "-" ) ? System.in : new FileInputStream( map ), 1024 * 1024 ) );
				for( int i = 0 ; i < f.length; i++ ) f[ i ] = dis.readInt();   
				dis.close();
			}
			System.err.println( " done." ); 
			ImmutableGraph g2 = map( graph[ 0 ], f, pm );
			System.err.println("Graph computed");
			ImmutableGraph.store( destGraphClass, g2, dest,  pm );
		} 
		else if ( transform.equals( "arcfilter" ) ) ImmutableGraph.store( destGraphClass, filterArcs( graph[ 0 ], arcFilter, pm ), dest, pm );
		else if ( transform.equals( "symmetrize" ) ) ImmutableGraph.store( destGraphClass, symmetrize( graph[ 0 ], graph[ 1 ], pm ), dest, pm );
		else if ( transform.equals( "transpose" ) ) ImmutableGraph.store( destGraphClass, transpose( graph[ 0 ], pm ), dest,  pm );
		else if ( transform.equals( "union" ) ) ImmutableGraph.store( destGraphClass, union( graph[ 0 ], graph[ 1 ], pm ), dest, pm );
		else if ( transform.equals( "gray" ) ) ImmutableGraph.store( destGraphClass, map( graph[ 0 ], grayCodePermutation( graph[ 0 ] ) ), dest, pm );
	}
}

// Local Variables:
// mode: jde
// tab-width: 4
// End:
