package it.unimi.dsi.webgraph;


/*		 
 * Copyright (C) 2003 Paolo Boldi and Sebastiano Vigna 
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

import gnu.getopt.Getopt;
import gnu.getopt.LongOpt;
import it.unimi.dsi.mg4j.index.CompressionFlags;
import it.unimi.dsi.mg4j.util.ProgressMeter;

import java.io.IOException;
import java.util.Arrays;

/** The main method of this class allows one to transpose a {@link BVGraph}, obtaining
 * a new {@link BVGraph}. The method used is not particularly smart, and requires
 * at least four bytes of memory per link.
 * 
 * @deprecated Use {@link it.unimi.dsi.webgraph.Transform#transpose(ImmutableGraph)} instead.
 */

public class TransposeBVGraph implements CompressionFlags {

	private TransposeBVGraph() {}

	private static void transposeBVGraph( String sourceBasename, String destBasename ) throws IOException {
		int i, j, d, a[];
		// Load the source BVGraph without actually loading it into memory.
		BVGraph source = (BVGraph)BVGraph.loadOffline( sourceBasename );

		final int n = source.n;
		final int numPred[] = new int[ n ];

		System.err.print( "Counting predecessors..." );
		ProgressMeter pm = new ProgressMeter( 1000000 );
		pm.itemsName( "nodes" );
		pm.start();
		NodeIterator nodeIterator = source.nodeIterator();
		i = n;
		while( i-- != 0 ) {
			nodeIterator.nextInt();
			d = nodeIterator.outdegree();
			a = nodeIterator.successorArray();
			while( d-- != 0 ) numPred[ a[ d ] ]++;
			pm.update();
		}
		pm.stop();
		System.err.println( " done." );
		System.err.println( pm );

		final int pred[][] = new int[ n ][];

		System.err.print( "Allocating memory for predecessors..." );
		i = n;
		while( i-- != 0 ) if ( numPred[ i ] != 0 ) pred[ i ] = new int[ numPred[ i ] ];
		System.err.println( " done." );

		i = n;
		while( i-- != 0 ) numPred[ i ] = 0;

		System.err.print( "Computing predecessors..." );
		pm.start();
		nodeIterator = source.nodeIterator();
		i = n;
		while( i-- != 0 ) {
			j = nodeIterator.nextInt();
			d = nodeIterator.outdegree();
			a = nodeIterator.successorArray();
			while( d-- != 0 ) pred[ a[ d ] ][ numPred[ a[ d ] ]++ ] = j;
			pm.update();
		}
		pm.stop();
		System.err.println( " done." );
		System.err.println( pm );

		i = n;
		System.err.print( "Sorting predecessors..." );
		pm.start();
		while( i-- != 0 ) {
			if ( pred[ i ] != null ) Arrays.sort( pred[ i ] );
			pm.update();
		} 
		pm.stop();
		System.err.println( " done." );
		System.err.println( pm );

		BVGraph.store( new ImmutableGraph() {
			private final int[] empty = new int[ 0 ];
			public int numNodes() { return n; }
			public int[] successorArray( int x ) { return pred[ x ] != null ? pred[ x ] : empty; }
			public int outdegree( int x ) { return numPred[ x ]; }
			},
			destBasename,
			pm
		);
	}

	public static void main( String args[] ) throws IOException {
		int ch;
		LongOpt longopts[] = new LongOpt[] {
			new LongOpt( "help", LongOpt.NO_ARGUMENT, null, 'h' ),
		};

		Getopt g = new Getopt( "TransposeBVGraph", args, "h", longopts );
		g.setOpterr( true );

		while ( ( ch = g.getopt() ) != -1 ) switch ( ch ) {
		case 'h':
			System.err.println( "Usage: TransposeBVGraph SOURCE DEST" );
			System.err.println( "Transposes the SOURCE BVGraph into the DEST BVGraph." );
			System.err.println( "Help:" );
			System.err.println( "  -h, --help              print this help screen" );
			System.err.println( "" );
			return;

		case '?':
			return;

		default:
			break;
		}

		if ( args.length - g.getOptind() != 2 ) {
			System.err.println( "Wrong number (" + ( args.length - g.getOptind() ) + ") of arguments." );
			return;
		}
		
		transposeBVGraph( args[ g.getOptind() ], args[ g.getOptind() + 1 ] );
	}
}

// Local Variables:
// mode: jde
// tab-width: 4
// End:
